import type { Database } from 'jsr:@db/sqlite';
import { encodeBase64 } from 'jsr:@std/encoding@0.224/base64';
import { hexToBytes } from '@noble/ciphers/utils';
import * as nostrTools from '@nostr/tools';
import type { UserConnection } from '../UserConnection.ts';
import { handleSocketError } from '../index.ts';
import { createNewCCN } from '../utils/createNewCCN.ts';
import { encryptUint8Array, encryptionKey } from '../utils/encryption.ts';
import { getEveFilePath } from '../utils/files.ts';
import { getAllCCNs } from '../utils/getAllCCNs.ts';
import { log } from '../utils/logs.ts';
import { sql } from '../utils/queries.ts';
import {
  SecurityEventType,
  SecuritySeverity,
  logAuthEvent,
  logSecurityEvent,
} from '../utils/securityLogs.ts';

function activateCCN(database: Database, pubkey: string): void {
  sql`UPDATE ccns SET is_active = 0`(database);
  sql`UPDATE ccns SET is_active = 1 WHERE pubkey = ${pubkey}`(database);

  logAuthEvent(SecurityEventType.CCN_ACTIVATION_ATTEMPT, true, {
    ccn_pubkey: pubkey,
  });
}

async function handleCreateCCN(
  connection: UserConnection,
  data: { name: string; seed?: string; creator: string },
): Promise<void> {
  log.debug('start', { tag: 'handleCreateCCN', data });

  logSecurityEvent({
    eventType: SecurityEventType.CCN_CREATION_ATTEMPT,
    severity: SecuritySeverity.MEDIUM,
    source: 'ccn_management',
    details: {
      ccn_name: data.name,
      creator: data.creator,
      has_seed: !!data.seed,
    },
  });

  try {
    if (!data.name || typeof data.name !== 'string') {
      logSecurityEvent({
        eventType: SecurityEventType.CCN_CREATION_ATTEMPT,
        severity: SecuritySeverity.MEDIUM,
        source: 'ccn_management',
        details: {
          error: 'invalid_name',
          name_provided: !!data.name,
          name_type: typeof data.name,
        },
      });

      connection.sendNotice('Name is required');
      return;
    }

    if (!data.creator || typeof data.creator !== 'string') {
      logSecurityEvent({
        eventType: SecurityEventType.CCN_CREATION_ATTEMPT,
        severity: SecuritySeverity.MEDIUM,
        source: 'ccn_management',
        details: {
          error: 'invalid_creator',
          creator_provided: !!data.creator,
          creator_type: typeof data.creator,
        },
      });

      connection.sendNotice('Creator is required');
      return;
    }

    const newCcn = await createNewCCN(
      connection.db,
      data.name,
      data.creator,
      data.seed,
    );
    log.debug('created new CCN', {
      tag: 'handleCreateCCN',
      pubkey: newCcn.pubkey,
    });
    activateCCN(connection.db, newCcn.pubkey);
    log.debug('activated new CCN', {
      tag: 'handleCreateCCN',
      pubkey: newCcn.pubkey,
    });

    logSecurityEvent({
      eventType: SecurityEventType.CCN_CREATION_ATTEMPT,
      severity: SecuritySeverity.LOW,
      source: 'ccn_management',
      details: {
        success: true,
        ccn_pubkey: newCcn.pubkey,
        ccn_name: data.name,
        creator: data.creator,
      },
    });

    connection.sendResponse([
      'OK',
      'CCN CREATED',
      true,
      JSON.stringify({
        pubkey: newCcn.pubkey,
        name: data.name,
      }),
    ]);

    log.info('CCN created', data);
  } catch (error: unknown) {
    log.error('error', { tag: 'handleCreateCCN', error });

    logSecurityEvent({
      eventType: SecurityEventType.CCN_CREATION_ATTEMPT,
      severity: SecuritySeverity.HIGH,
      source: 'ccn_management',
      details: {
        success: false,
        error_message: error instanceof Error ? error.message : 'Unknown error',
        ccn_name: data.name,
        creator: data.creator,
      },
    });

    handleSocketError(connection, 'create CCN', error);
  }
  log.debug('end', { tag: 'handleCreateCCN' });
}

function handleGetCCNs(connection: UserConnection): void {
  try {
    const ccns = getAllCCNs(connection.db);
    connection.sendResponse(['OK', 'CCN LIST', true, JSON.stringify(ccns)]);
  } catch (error: unknown) {
    handleSocketError(connection, 'get CCNs', error);
  }
}

function handleActivateCCN(
  connection: UserConnection,
  data: { pubkey: string },
): void {
  log.debug('start', { tag: 'handleActivateCCN', data });
  try {
    if (!data.pubkey || typeof data.pubkey !== 'string') {
      connection.sendNotice('CCN pubkey is required');
      return;
    }

    const ccnExists = sql`
      SELECT COUNT(*) as count FROM ccns WHERE pubkey = ${data.pubkey}
    `(connection.db)[0].count;

    if (ccnExists === 0) {
      connection.sendNotice('CCN not found');
      log.debug('CCN not found', {
        tag: 'handleActivateCCN',
        pubkey: data.pubkey,
      });
      return;
    }

    for (const subscriptionId of connection.subscriptions.keys()) {
      connection.sendResponse([
        'CLOSED',
        subscriptionId,
        'Subscription closed due to CCN activation',
      ]);
      log.debug('closed subscription', {
        tag: 'handleActivateCCN',
        subscriptionId,
      });
    }

    connection.subscriptions.clear();
    log.info('All subscriptions cleared due to CCN activation', {});

    activateCCN(connection.db, data.pubkey);
    log.debug('activated CCN', {
      tag: 'handleActivateCCN',
      pubkey: data.pubkey,
    });

    const activatedCCN = sql`
      SELECT pubkey, name FROM ccns WHERE pubkey = ${data.pubkey}
    `(connection.db)[0];

    connection.sendResponse([
      'OK',
      'CCN ACTIVATED',
      true,
      JSON.stringify(activatedCCN),
    ]);

    log.info(`CCN activated: ${activatedCCN.name}`, {});
  } catch (error: unknown) {
    log.error('error', { tag: 'handleActivateCCN', error });
    handleSocketError(connection, 'activate CCN', error);
  }
  log.debug('end', { tag: 'handleActivateCCN' });
}

async function handleAddCCN(
  connection: UserConnection,
  data: { name: string; allowedPubkeys: string[]; privateKey: string },
): Promise<void> {
  log.debug('start', { tag: 'handleAddCCN', data });
  try {
    if (!data.privateKey || typeof data.privateKey !== 'string') {
      connection.sendNotice('CCN private key is required');
      return;
    }

    const privateKeyBytes = hexToBytes(data.privateKey);
    const pubkey = nostrTools.getPublicKey(privateKeyBytes);
    log.debug('derived pubkey', { tag: 'handleAddCCN', pubkey });

    const ccnExists = sql`
      SELECT COUNT(*) as count FROM ccns WHERE pubkey = ${pubkey}
    `(connection.db)[0].count;

    if (ccnExists > 0) {
      connection.sendNotice('CCN already exists');
      log.debug('CCN already exists', {
        tag: 'handleAddCCN',
        pubkey,
      });
      return;
    }

    const ccnPublicKey = nostrTools.getPublicKey(privateKeyBytes);
    const ccnPrivPath = await getEveFilePath(`ccn_keys/${ccnPublicKey}`);
    const encryptedPrivateKey = encryptUint8Array(
      privateKeyBytes,
      encryptionKey,
    );
    Deno.writeTextFileSync(ccnPrivPath, encodeBase64(encryptedPrivateKey));

    connection.db.run('BEGIN TRANSACTION');
    log.debug('begin transaction', { tag: 'handleAddCCN' });

    sql`INSERT INTO ccns (pubkey, name) VALUES (${ccnPublicKey}, ${data.name})`(
      connection.db,
    );
    for (const allowedPubkey of data.allowedPubkeys)
      sql`INSERT INTO allowed_writes (ccn_pubkey, pubkey) VALUES (${ccnPublicKey}, ${allowedPubkey})`(
        connection.db,
      );

    connection.db.run('COMMIT TRANSACTION');
    log.debug('committed transaction', { tag: 'handleAddCCN' });
    activateCCN(connection.db, ccnPublicKey);
    log.debug('activated CCN', {
      tag: 'handleAddCCN',
      pubkey: ccnPublicKey,
    });

    connection.sendResponse([
      'OK',
      'CCN ADDED',
      true,
      JSON.stringify({
        pubkey: ccnPublicKey,
        name: 'New CCN',
      }),
    ]);
  } catch (error: unknown) {
    log.error('error', { tag: 'handleAddCCN', error });
    handleSocketError(connection, 'ADD CCN', error);
  }
  log.debug('end', { tag: 'handleAddCCN' });
}

function handleCCNCommands(
  connection: UserConnection,
  command: string,
  data: unknown,
) {
  switch (command) {
    case 'CREATE':
      return handleCreateCCN(
        connection,
        data as { name: string; seed?: string; creator: string },
      );
    case 'ADD':
      return handleAddCCN(
        connection,
        data as { name: string; allowedPubkeys: string[]; privateKey: string },
      );
    case 'LIST':
      return handleGetCCNs(connection);
    case 'ACTIVATE':
      return handleActivateCCN(connection, data as { pubkey: string });
    default:
      return log.warn('Invalid CCN command', {});
  }
}

export {
  activateCCN,
  handleActivateCCN,
  handleAddCCN,
  handleCCNCommands,
  handleCreateCCN,
  handleGetCCNs,
};
