import * as nostrTools from '@nostr/tools';
import type { UserConnection } from '../UserConnection.ts';
import { addEventToDb } from '../dbEvents/addEventToDb.ts';
import {
  EventAlreadyExistsException,
  createEncryptedEvent,
} from '../eventEncryptionDecryption.ts';
import { filtersMatchingEvent } from '../utils/filtersMatchingEvent.ts';
import { getActiveCCN } from '../utils/getActiveCCN.ts';
import { isArray } from '../utils/isArray.ts';
import { log } from '../utils/logs.ts';
import { queueEventForTransmission } from '../utils/outboundQueue.ts';

export async function handleEvent(
  connection: UserConnection,
  event: nostrTools.Event,
) {
  log.debug('start', { tag: 'handleEvent', eventId: event.id });
  const valid = nostrTools.verifyEvent(event);
  if (!valid) {
    connection.sendNotice('Invalid event');
    return log.warn('Invalid event', { tag: 'handleEvent' });
  }

  const activeCCN = getActiveCCN(connection.db);
  if (!activeCCN) {
    connection.sendNotice('No active CCN found');
    return log.warn('No active CCN found', { tag: 'handleEvent' });
  }

  const encryptedEvent = await createEncryptedEvent(event, connection.db);
  try {
    if (isArray(encryptedEvent)) {
      log.debug('adding chunked event to database', {
        tag: 'handleEvent',
      });
      addEventToDb(connection.db, event, encryptedEvent[0], activeCCN.pubkey);
    } else {
      addEventToDb(connection.db, event, encryptedEvent, activeCCN.pubkey);
    }

    queueEventForTransmission(
      connection.db,
      event.id,
      encryptedEvent,
      activeCCN.pubkey,
    );

    log.debug('event queued for transmission', {
      tag: 'handleEvent',
      eventId: event.id,
    });
  } catch (e) {
    if (e instanceof EventAlreadyExistsException) {
      log.warn('Event already exists');
      return;
    }
    if (e instanceof Error)
      log.error('error adding event', {
        tag: 'handleEvent',
        error: e.stack,
      });
    else
      log.error('error adding event', {
        tag: 'handleEvent',
        error: String(e),
      });
  }

  connection.sendOK(event.id, true, 'Event added');
  log.debug('sent OK', { tag: 'handleEvent', eventId: event.id });

  const filtersThatMatchEvent = filtersMatchingEvent(event, connection);

  for (let i = 0; i < filtersThatMatchEvent.length; i++) {
    const filter = filtersThatMatchEvent[i];
    connection.sendEvent(filter, event);
    log.debug('sent event to filter', {
      tag: 'handleEvent',
      filter,
      eventId: event.id,
    });
  }
  log.debug('end', { tag: 'handleEvent', eventId: event.id });
}
