import type { NostrClientREQ } from 'jsr:@nostrify/types';
import type { UserConnection } from '../UserConnection.ts';
import { isCCNReplaceableEvent } from '../utils/eventTypes.ts';
import { getActiveCCN } from '../utils/getActiveCCN.ts';
import { log } from '../utils/logs.ts';
import { parseATagQuery } from '../utils/parseATagQuery.ts';
import { mixQuery, sql, sqlPartial } from '../utils/queries.ts';

export function handleRequest(
  connection: UserConnection,
  request: NostrClientREQ,
) {
  log.debug('start', { tag: 'handleRequest', request });
  const [, subscriptionId, ...filters] = request;
  if (connection.subscriptions.has(subscriptionId)) {
    return log.warn('Duplicate subscription ID', {
      tag: 'handleRequest',
    });
  }

  log.info(
    `New subscription: ${subscriptionId} with filters: ${JSON.stringify(
      filters,
    )}`,
  );

  const activeCCN = getActiveCCN(connection.db);
  if (!activeCCN) {
    connection.sendNotice('No active CCN found');
    return log.warn('No active CCN found', { tag: 'handleRequest' });
  }

  let query = sqlPartial`SELECT * FROM events WHERE replaced = 0 AND deleted = 0 AND ccn_pubkey = ${activeCCN.pubkey}`;

  let minLimit: number | null = null;
  for (const filter of filters) {
    if (filter.limit && filter.limit > 0) {
      minLimit =
        minLimit === null ? filter.limit : Math.min(minLimit, filter.limit);
    }
  }

  const filtersAreNotEmpty = filters.some((filter) => {
    return Object.values(filter).some((value) => {
      return value.length > 0;
    });
  });

  if (filtersAreNotEmpty) {
    query = mixQuery(query, sqlPartial`AND`);

    for (let i = 0; i < filters.length; i++) {
      // filters act as OR, filter groups act as AND
      query = mixQuery(query, sqlPartial`(`);

      const filter = Object.entries(filters[i]).filter(([type, value]) => {
        if (type === 'ids') return value.length > 0;
        if (type === 'authors') return value.length > 0;
        if (type === 'kinds') return value.length > 0;
        if (type.startsWith('#')) return value.length > 0;
        if (type === 'since') return value > 0;
        if (type === 'until') return value > 0;
        if (type === 'limit') return value > 0;
        return false;
      });

      const filterWithoutLimit = filter.filter(([type]) => type !== 'limit');

      for (let j = 0; j < filter.length; j++) {
        const [type, value] = filter[j];

        if (type === 'ids') {
          const uniqueIds = [...new Set(value)];
          query = mixQuery(query, sqlPartial`(`);
          for (let k = 0; k < uniqueIds.length; k++) {
            const id = uniqueIds[k] as string;

            query = mixQuery(query, sqlPartial`(id = ${id})`);

            if (k < uniqueIds.length - 1) {
              query = mixQuery(query, sqlPartial`OR`);
            }
          }
          query = mixQuery(query, sqlPartial`)`);
        }

        if (type === 'authors') {
          const uniqueAuthors = [...new Set(value)];
          query = mixQuery(query, sqlPartial`(`);
          for (let k = 0; k < uniqueAuthors.length; k++) {
            const author = uniqueAuthors[k] as string;

            query = mixQuery(query, sqlPartial`(pubkey = ${author})`);

            if (k < uniqueAuthors.length - 1) {
              query = mixQuery(query, sqlPartial`OR`);
            }
          }
          query = mixQuery(query, sqlPartial`)`);
        }

        if (type === 'kinds') {
          const uniqueKinds = [...new Set(value)];
          query = mixQuery(query, sqlPartial`(`);
          for (let k = 0; k < uniqueKinds.length; k++) {
            const kind = uniqueKinds[k] as number;

            query = mixQuery(query, sqlPartial`(kind = ${kind})`);

            if (k < uniqueKinds.length - 1) {
              query = mixQuery(query, sqlPartial`OR`);
            }
          }
          query = mixQuery(query, sqlPartial`)`);
        }

        if (type.startsWith('#')) {
          const tag = type.slice(1);
          const uniqueValues = [...new Set(value)];
          query = mixQuery(query, sqlPartial`(`);
          for (let k = 0; k < uniqueValues.length; k++) {
            const tagValue = uniqueValues[k] as string;
            if (tag === 'a') {
              const aTagInfo = parseATagQuery(tagValue);

              if (aTagInfo.dTag && aTagInfo.dTag !== '') {
                if (isCCNReplaceableEvent(aTagInfo.kind)) {
                  // CCN replaceable event reference
                  query = mixQuery(
                    query,
                    sqlPartial`id IN (
                      SELECT e.id
                      FROM events e
                      JOIN event_tags t ON e.id = t.event_id
                      JOIN event_tags_values v ON t.tag_id = v.tag_id
                      WHERE e.kind = ${aTagInfo.kind}
                      AND t.tag_name = 'd'
                      AND v.value_position = 1
                      AND v.value = ${aTagInfo.dTag}
                    )`,
                  );
                } else {
                  // Addressable event reference
                  query = mixQuery(
                    query,
                    sqlPartial`id IN (
                      SELECT e.id
                      FROM events e
                      JOIN event_tags t ON e.id = t.event_id
                      JOIN event_tags_values v ON t.tag_id = v.tag_id
                      WHERE e.kind = ${aTagInfo.kind}
                      AND e.pubkey = ${aTagInfo.pubkey}
                      AND t.tag_name = 'd'
                      AND v.value_position = 1
                      AND v.value = ${aTagInfo.dTag}
                    )`,
                  );
                }
              } else {
                // Replaceable event reference
                query = mixQuery(
                  query,
                  sqlPartial`id IN (
                    SELECT id
                    FROM events
                    WHERE kind = ${aTagInfo.kind}
                    AND pubkey = ${aTagInfo.pubkey}
                  )`,
                );
              }
            } else {
              // Regular tag handling (unchanged)
              query = mixQuery(
                query,
                sqlPartial`id IN (
                  SELECT t.event_id
                  FROM event_tags t
                  WHERE t.tag_name = ${tag}
                  AND t.tag_id IN (
                    SELECT v.tag_id
                    FROM event_tags_values v
                    WHERE v.value_position = 1
                    AND v.value = ${tagValue}
                  )
                )`,
              );
            }

            if (k < uniqueValues.length - 1) {
              query = mixQuery(query, sqlPartial`OR`);
            }
          }
          query = mixQuery(query, sqlPartial`)`);
        }

        if (type === 'since') {
          query = mixQuery(query, sqlPartial`created_at > ${value}`);
        }

        if (type === 'until') {
          query = mixQuery(query, sqlPartial`created_at <= ${value}`);
        }

        if (j < filterWithoutLimit.length - 1)
          query = mixQuery(query, sqlPartial`AND`);
      }

      query = mixQuery(query, sqlPartial`)`);

      if (i < filters.length - 1) query = mixQuery(query, sqlPartial`OR`);
    }
  }

  query = mixQuery(query, sqlPartial`ORDER BY created_at ASC`);

  if (minLimit !== null) {
    query = mixQuery(query, sqlPartial`LIMIT ${minLimit}`);
  }

  log.debug('built query', {
    tag: 'handleRequest',
    query: query.query,
    values: query.values,
  });

  const events = connection.db.prepare(query.query).all(...query.values);
  log.debug('found events', {
    tag: 'handleRequest',
    count: events.length,
  });

  for (let i = 0; i < events.length; i++) {
    const rawTags = sql`SELECT * FROM event_tags_view WHERE event_id = ${
      events[i].id
    }`(connection.db);
    const tagsByIndex = new Map<
      number,
      {
        name: string;
        values: Map<number, string>;
      }
    >();

    for (const tag of rawTags) {
      let tagData = tagsByIndex.get(tag.tag_index);
      if (!tagData) {
        tagData = {
          name: tag.tag_name,
          values: new Map(),
        };
        tagsByIndex.set(tag.tag_index, tagData);
      }

      tagData.values.set(tag.tag_value_position, tag.tag_value);
    }

    const tagsArray = Array.from(tagsByIndex.entries())
      .sort(([indexA], [indexB]) => indexA - indexB)
      .map(([_, tagData]) => {
        const { name, values } = tagData;

        return [
          name,
          ...Array.from(values.entries())
            .sort(([posA], [posB]) => posA - posB)
            .map(([_, value]) => value),
        ];
      });

    const event = {
      id: events[i].id,
      pubkey: events[i].pubkey,
      created_at: events[i].created_at,
      kind: events[i].kind,
      tags: tagsArray,
      content: events[i].content,
      sig: events[i].sig,
    };

    connection.sendEvent(subscriptionId, event);
    log.debug('sent event', {
      tag: 'handleRequest',
      subscriptionId,
      eventId: event.id,
    });
  }
  connection.sendEOSE(subscriptionId);
  log.debug('sent EOSE', { tag: 'handleRequest', subscriptionId });
  connection.subscriptions.set(subscriptionId, filters);
  log.debug('end', { tag: 'handleRequest', subscriptionId });
}
